from fastapi import FastAPI
from fastapi.testclient import TestClient
from pytest import fixture

from fastapi_restful import Resource, Api

app = FastAPI()


@fixture
def resource_for_test():
    class TestResource(Resource):
        pass
    return TestResource


@fixture
def fastapi_test_client():
    return TestClient(app)


@fixture
def api_for_test():
    return Api(app)



from typing import Type

from fastapi import FastAPI

from .resource import Resource


class Api:
    """A class for create RESTfull-API."""

    fastapi: FastAPI = None

    def __init__(self, fastapi_app: FastAPI = None):
        if fastapi_app is not None:
            self.init_fastapi(fastapi_app)

    def init_fastapi(self, fastapi_app: FastAPI) -> None:
        """
        Will connect the FastAPI application to this Api instance.

        Parameters
        ----------
        fastapi_app: FastAPI application instance.

        Returns
        -------
        None
        """
        self.fastapi = fastapi_app

    def add_resource(self, resource: Type[Resource], path: str = None) -> None:
        """
        Initial resource and add resource router to FastAPI.

        Parameters
        ----------
        resource: Type of Resource
        path: Resource path

        Returns
        -------
        None
        """
        res = resource(path)
        self.fastapi.include_router(res.router)



from inspect import getmembers, ismethod, Parameter, Signature, signature
from typing import List, Callable, Tuple, Dict, Any

from fastapi import APIRouter, HTTPException, params

from .utils import is_overridden_func


class ManageSignature:
    """A class for interacting with a function signature."""

    _required_args: List[str]

    @staticmethod
    def __get_kwargs__(func: Callable) -> Dict[str, Any]:
        """
        Get keyword names and default values from func.
        Skip item If default value is Query instance of FastAPI.

        Parameters
        ----------
        func: Target func

        Returns
        -------
        Kwargs
        """
        sign = signature(func)
        return {
            k: v.default
            for k, v in sign.parameters.items()
            if v.default is not Parameter.empty and not isinstance(v.default, params.Query)
        }

    def __replace_signature__(self, func: Callable) -> Callable:
        """
        Decorator for set new signature.

        Parameters
        ----------
        func: Target func

        Returns
        -------
        New function with new signature.
        """

        def new_func(*args, **kwargs):
            return func(*args, **kwargs)
        sign = signature(func)
        new_params = self.__gen_new_params__(sign)
        new_func.__signature__ = sign.replace(parameters=new_params)
        return new_func

    def __gen_new_params__(self, sign: Signature) -> List[Parameter]:
        """
        Generate list with new params for signature.

        Parameters
        ----------
        sign: Signature

        Returns
        -------
        List with new params.
        """
        old_params = list(sign.parameters.values())
        new_params = [p.replace(kind=Parameter.POSITIONAL_OR_KEYWORD) for p in old_params
                      if p.name in self._required_args or isinstance(p.default, params.Query)]
        return new_params


class HTTPMethods:
    """The supported HTTP methods are described here."""

    _HTTP_METHODS: tuple = ('head', 'options', 'get', 'post', 'patch', 'put', 'delete')

    def _get_route_handlers(self) -> List[Tuple[str, Callable]]:
        """
        Get route handler functions.

        Returns
        -------
        List with data o route handlers.
        """
        filter_obj = filter(
            lambda x: x[0] in self._HTTP_METHODS and is_overridden_func(x[1]), getmembers(self, predicate=ismethod)
        )
        return list(filter_obj)

    def head(self):
        raise HTTPException(status_code=405, detail="Method Not Allowed")

    def options(self):
        raise HTTPException(status_code=405, detail="Method Not Allowed")

    def get(self, **kwargs):
        raise HTTPException(status_code=405, detail="Method Not Allowed")

    def post(self, **kwargs):
        raise HTTPException(status_code=405, detail="Method Not Allowed")

    def patch(self):
        raise HTTPException(status_code=405, detail="Method Not Allowed")

    def put(self):
        raise HTTPException(status_code=405, detail="Method Not Allowed")

    def delete(self):
        raise HTTPException(status_code=405, detail="Method Not Allowed")


class Resource(HTTPMethods, ManageSignature):
    """This class is designed to implement CRUD methods of a specific api resource with FastAPI."""

    router: APIRouter
    tag: str = None
    path: str = ''

    def __init__(self, path: str = None):
        """
        Init Resource instance.
        If the value of the 'tag' attribute is not set. Then the value will be the name of the class instance.

        Parameters
        ----------
        path: Path for this resource
        """
        if self.tag is None:
            self.tag = self.__class__.__name__
        if path is not None:
            self.path = path
        self.__init_router__()

    def __init_router__(self) -> None:
        """
        Init APIRouter of FastAPI and include implemented methods.

        Returns
        -------
        None
        """
        self.router = APIRouter(prefix=self.path, tags=[self.tag])
        self.__include_methods__()

    def __include_methods__(self):
        """
        Include HTTP methods to router.

        Get kwargs from implemented method (only for route), and replace signature of route_handler,
        then add api route to router.

        Returns
        -------
        None
        """
        for method, handler in self._get_route_handlers():
            kwargs = self.__get_kwargs__(handler)
            route_handler = self.__replace_signature__(handler)
            self.router.add_api_route(path='', endpoint=route_handler, methods=[method], **kwargs)

    @property
    def _required_args(self) -> List[str]:
        """
        Get required arg names from path of this Resource instance.

        Returns
        -------
        List with required arg names.
        """
        return [item.strip('{}') for item in self.path.split('/') if item.startswith('{') and item.endswith('}')]



from fastapi import FastAPI

from fastapi_restful import Api


def test_init_api_with_fastapi():
    app = FastAPI()
    api = Api(fastapi_app=app)
    assert api.fastapi is app


def test_init_api_without_fastapi():
    api = Api()
    assert api.fastapi is None


def test_init_fastapi():
    app = FastAPI()
    api = Api()
    api.init_fastapi(app)
    assert api.fastapi is app


def test_add_resource(resource_for_test, api_for_test):
    api_for_test.add_resource(resource_for_test)



import pytest
from fastapi import APIRouter, HTTPException


def test_init_resource_default_tag(resource_for_test):
    instance = resource_for_test()
    assert instance.tag == resource_for_test.__name__


def test_init_resource_custom_tag(resource_for_test):
    custom_tag = 'Test custom tag'
    resource_for_test.tag = custom_tag
    instance = resource_for_test()
    assert instance.tag == custom_tag


def test_init_resource_default_path(resource_for_test):
    instance = resource_for_test()
    assert instance.path == ''


def test_init_resource_custom_path(resource_for_test):
    custom_path = '/my/path'
    instance = resource_for_test(path=custom_path)
    assert instance.path == custom_path


def test_init_resource_init_router(resource_for_test):
    instance = resource_for_test()
    assert instance.router is not None
    assert isinstance(instance.router, APIRouter)
    assert instance.router.tags == [resource_for_test.__name__]
    assert instance.router.prefix == resource_for_test.path


def test_init_resource_include_methods(resource_for_test):
    def get(self, **kwargs):
        return {}

    instance_before = resource_for_test()
    assert len(instance_before.router.routes) == 0
    resource_for_test.get = get
    instance = resource_for_test()
    assert len(instance.router.routes) == 1


def test_resource_required_args_in_path(resource_for_test):
    instance = resource_for_test(path='/api/{path_arg}')
    assert instance._required_args == ['path_arg']


def test_not_implemented_method(resource_for_test):
    instance = resource_for_test()
    for method in instance._HTTP_METHODS:
        func = getattr(instance, method)
        pytest.raises(HTTPException, func)


def test_execute_func_with_new_signature(resource_for_test):
    output = {}

    def get(self, **kwargs):
        return output

    resource_for_test.get = get
    instance = resource_for_test()
    new_func = instance.__replace_signature__(instance.get)
    assert new_func() == output



def is_overridden_func(target_func) -> bool:
    """
    Check if the function is overridden in child class.

    Parameters
    ----------
    target_func: Target function

    Returns
    -------
    True/False
    """
    child_instance = target_func.__self__
    func_of_parent = getattr(super(type(child_instance), child_instance), target_func.__name__)
    return target_func.__func__ != func_of_parent.__func__



from setuptools import setup, find_packages

setup(
    name='fastapi-restful-extension',
    version='0.1.0',
    author="Shumilo Maxim",
    author_email="shumilo.mk@gmail.com",
    description='Extension for make RESTfull interfaces with FastAPI.',
    long_description="",
    install_requires=['fastapi~=0.78.0', 'setuptools~=62.3.2'],
    packages=find_packages(),
    include_package_data=True,
    url='https://github.com/maximshumilo/fastapi-restful-extension',
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6'
)
